/*
 * Copyright (C) 2023 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *	  Author: Suvendhu Hansa <suvendhu.hansa@iopsys.eu>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <uci.h>
#include <libubox/uloop.h>
#include <libubus.h>

#include "helper.h"

extern struct ubus_context *g_ubus_ctx;

static int timemngr_uci_init(void);
static int timemngr_uci_fini(void);
static bool timemngr_uci_validate_section(const char *str);
static int timemngr_uci_init_ptr(struct uci_ptr *ptr, const char *package, const char *section, const char *option, const char *value);
static struct uci_section *timemngr_uci_walk_section(const char *package, const char *section_type, struct uci_section *prev_section);
static struct uci_element *timemngr_uci_lookup_list(struct uci_list *list, const char *name);
static int timemngr_uci_lookup_ptr_by_section(struct uci_ptr *ptr, struct uci_section *section, const char *option, const char *value);
static char *timemngr_uci_get_value_by_section(struct uci_section *section, const char *option);
struct uloop_timeout timemngr_sync_timer = { .cb = timemngr_schedule_sync };

#define timemngr_uci_foreach_section_safe(package, section_type, _tmp, section) \
	for(section = timemngr_uci_walk_section(package, section_type, NULL), \
		_tmp = (section) ? timemngr_uci_walk_section(package, section_type, section) : NULL; \
		section != NULL; \
		section = _tmp, _tmp = (section) ? timemngr_uci_walk_section(package, section_type, section) : NULL)

static bool get_timemngr_enable(void);
static bool get_timemngr_server_enable(void);
static bool get_timemngr_client_enable(const char *sec);
static bool ntpd_running(void);
static bool ntpd_leap_alarm(void);
static void add_counters_to_blob(struct blob_buf *bb, unsigned int pack_sent,
		unsigned int pack_sent_fail, unsigned int pack_recv,
		unsigned int pack_drop);
static void update_ntpd_counters(struct blob_buf *bb);

static struct uci_context *uci_ctx_timemngr = NULL;

static int timemngr_uci_init(void)
{
	uci_ctx_timemngr = uci_alloc_context();
	if (!uci_ctx_timemngr)
		return -1;

	uci_set_confdir(uci_ctx_timemngr, "/etc/config/");

	return 0;
}

static int timemngr_uci_fini(void)
{
	/* Freed uci context  */
	if (uci_ctx_timemngr)
		uci_free_context(uci_ctx_timemngr);

	return 0;
}

static bool timemngr_uci_validate_section(const char *str)
{
	if (!*str)
		return false;

	for (; *str; str++) {
		unsigned char c = *str;

		if (isalnum(c) || c == '_')
			continue;

		return false;
	}

	return true;
}

static int timemngr_uci_init_ptr(struct uci_ptr *ptr, const char *package, const char *section, const char *option, const char *value)
{
	memset(ptr, 0, sizeof(struct uci_ptr));

	/* value */
	if (value)
		ptr->value = value;

	ptr->package = package;
	if (!ptr->package)
		goto error;

	ptr->section = section;
	if (!ptr->section) {
		ptr->target = UCI_TYPE_PACKAGE;
		goto lastval;
	}

	ptr->option = option;
	if (!ptr->option) {
		ptr->target = UCI_TYPE_SECTION;
		goto lastval;
	} else
		ptr->target = UCI_TYPE_OPTION;

lastval:
	if (ptr->section && !timemngr_uci_validate_section(ptr->section))
		ptr->flags |= UCI_LOOKUP_EXTENDED;

	return 0;

error:
	return -1;
}

static struct uci_section *timemngr_uci_walk_section(const char *package, const char *section_type, struct uci_section *prev_section)
{
	struct uci_ptr ptr;
	struct uci_element *e;
	struct uci_section *next_section;

	if (section_type == NULL) {
		if (prev_section) {
			e = &prev_section->e;
			if (e->list.next == &prev_section->package->sections)
				return NULL;
			e = container_of(e->list.next, struct uci_element, list);
			next_section = uci_to_section(e);
			return next_section;
		} else {
			if (timemngr_uci_init_ptr(&ptr, package, NULL, NULL, NULL))
				return NULL;

			if (uci_lookup_ptr(uci_ctx_timemngr, &ptr, NULL, true) != UCI_OK)
				return NULL;

			if (ptr.p->sections.next == &ptr.p->sections)
				return NULL;
			e = container_of(ptr.p->sections.next, struct uci_element, list);
			next_section = uci_to_section(e);

			return next_section;
		}
	} else {
		struct uci_list *ul = NULL, *shead = NULL;

		if (prev_section) {
			ul = &prev_section->e.list;
			shead = &prev_section->package->sections;
		} else {
			if (timemngr_uci_init_ptr(&ptr, package, NULL, NULL, NULL))
				return NULL;

			if (uci_lookup_ptr(uci_ctx_timemngr, &ptr, NULL, true) != UCI_OK)
				return NULL;

			ul = &ptr.p->sections;
			shead = &ptr.p->sections;
		}
		while (ul->next != shead) {
			e = container_of(ul->next, struct uci_element, list);
			next_section = uci_to_section(e);
			if (strcmp(next_section->type, section_type) == 0)
				return next_section;
			ul = ul->next;
		}
		return NULL;
	}
	return NULL;
}

static struct uci_element *timemngr_uci_lookup_list(struct uci_list *list, const char *name)
{
	struct uci_element *e;

	uci_foreach_element(list, e) {
		if (!strcmp(e->name, name))
			return e;
	}

	return NULL;
}

static int timemngr_uci_lookup_ptr_by_section(struct uci_ptr *ptr, struct uci_section *section, const char *option, const char *value)
{
	struct uci_element *e;
	memset(ptr, 0, sizeof(struct uci_ptr));

	ptr->package = section->package->e.name;
	ptr->section = section->e.name;
	ptr->option = option;
	ptr->value = value;
	ptr->flags |= UCI_LOOKUP_DONE;

	ptr->p = section->package;
	ptr->s = section;

	if (ptr->option) {
		e = timemngr_uci_lookup_list(&ptr->s->options, ptr->option);
		if (!e)
			return UCI_OK;
		ptr->o = uci_to_option(e);
		ptr->last = e;
		ptr->target = UCI_TYPE_OPTION;
	} else {
		ptr->last = &ptr->s->e;
		ptr->target = UCI_TYPE_SECTION;
	}

	ptr->flags |= UCI_LOOKUP_COMPLETE;

	return UCI_OK;
}

static char *timemngr_uci_get_value_by_section(struct uci_section *section, const char *option)
{
	struct uci_ptr ptr;

	if (timemngr_uci_lookup_ptr_by_section(&ptr, section, option, NULL) != UCI_OK)
		return "";

	if (!ptr.o)
		return "";

	if (ptr.o->v.string)
		return ptr.o->v.string;
	else
		return "";
}

static bool get_timemngr_enable(void)
{
	char *enable = "";
	if (timemngr_uci_init() != 0)
		return false;
	
	struct uci_section *s = NULL, *stmp = NULL;
	timemngr_uci_foreach_section_safe("time", "global", stmp, s) {
		enable = timemngr_uci_get_value_by_section(s, "enable");
		break;
	}

	timemngr_uci_fini();

	if (strlen(enable) == 0 || strcmp(enable, "1") == 0)
		return true;
	
	return false;
}

static bool get_timemngr_server_enable(void)
{
	char *enable = "";
	if (timemngr_uci_init() != 0)
		return false;
	
	struct uci_section *s = NULL, *stmp = NULL;
	timemngr_uci_foreach_section_safe("time", "server", stmp, s) {
		enable = timemngr_uci_get_value_by_section(s, "enable");
		break;
	}

	timemngr_uci_fini();

	if (strlen(enable) == 0 || strcmp(enable, "1") == 0)
		return true;
	
	return false;
}

static bool get_timemngr_client_enable(const char *sec)
{
	bool enable = false;

	if (timemngr_uci_init() != 0)
		return enable;
	
	struct uci_section *s = NULL, *stmp = NULL;
	timemngr_uci_foreach_section_safe("time", "client", stmp, s) {
		char *value = timemngr_uci_get_value_by_section(s, "enable");
		if (strlen(value) == 0 || strcmp(value, "1") == 0) {
			if (sec == NULL || strcmp(sec, s->e.name) == 0) {
				enable = true;
				break;
			}
		}
	}

	timemngr_uci_fini();

	return enable;
}

static bool ntpd_running(void)
{
	bool running = false;
	char buff[256] = {0};
	fd_set fdset;
	struct timeval timeout;

	timeout.tv_sec = 2;
	timeout.tv_usec = 0;

	FILE *fp = popen("pgrep /sbin/ntpd", "r");
	if (!fp) {
		PRINT_ERR("popen failed");
		return running;
	}

	int fd = fileno(fp);
	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);

	int rc = select(fd+1, &fdset, 0, 0, &timeout);
	if (rc == 0) {
		PRINT_DEBUG("select timedout");
		pclose(fp);
		return running;
	}

	if (NULL != fgets(buff, sizeof(buff), fp) && strlen(buff) != 0) {
		running = true;
	}

	pclose(fp);

	return running;
}

static bool ntpd_leap_alarm(void)
{
	bool alarm = true;
	char buff[256] = {0};
	fd_set fdset;
	struct timeval timeout;

	timeout.tv_sec = 2;
	timeout.tv_usec = 0;

	FILE *fp = popen("ntpq -c \'rv 0 stratum\'", "r");
	if (!fp) {
		PRINT_ERR("popen failed");
		return alarm;
	}

	int fd = fileno(fp);
	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);

	int rc = select(fd+1, &fdset, 0, 0, &timeout);
	if (rc == 0) {
		PRINT_DEBUG("select timedout");
		pclose(fp);
		return alarm;
	}

	if (NULL != fgets(buff, sizeof(buff), fp)) {
		PRINT_DEBUG("stratum: (%d) (%s)", strncmp(buff, "stratum=16", 10), buff);
		if (strlen(buff) && strncmp(buff, "stratum=16", 10) != 0)
			alarm = false;
	}

	pclose(fp);
	return alarm;
}

static void add_counters_to_blob(struct blob_buf *bb, unsigned int pack_sent,
		unsigned int pack_sent_fail, unsigned int pack_recv,
		unsigned int pack_drop)
{
	if (bb) {
		blobmsg_add_u64(bb, "PacketsSent", pack_sent);
		blobmsg_add_u64(bb, "PacketsSentFail", pack_sent_fail);
		blobmsg_add_u64(bb, "PacketsReceived", pack_recv);
		blobmsg_add_u64(bb, "PacketsDropped", pack_drop);
	}
}

static void update_ntpd_counters(struct blob_buf *bb)
{
	fd_set fdset;
	struct timeval timeout;
	unsigned int pack_sent = 0, pack_sent_fail = 0, pack_recv = 0, pack_drop = 0;
	char buff[256] = {0};

	if (!bb) {
		PRINT_ERR("blob is Null");
		return;
	}

	timeout.tv_sec = 2;
	timeout.tv_usec = 0;

	FILE *fp = popen("ntpq -c iostats", "r");
	if (!fp) {
		PRINT_ERR("popen failed");
		add_counters_to_blob(bb, 0, 0, 0, 0);
		return;
	}

	int fd = fileno(fp);
	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);

	int rc = select(fd+1, &fdset, 0, 0, &timeout);
	if (rc == 0) {
		PRINT_DEBUG("select timedout");
		add_counters_to_blob(bb, 0, 0, 0, 0);
		pclose(fp);
		return;
	}

	while (NULL != fgets(buff, sizeof(buff), fp)) {
		if (strstr(buff, "packets sent:") != NULL) {
			sscanf(buff, "packets sent: %u", &pack_sent);
		}

		if (strstr(buff, "packet send failures:") != NULL) {
			sscanf(buff, "packet send failures: %u", &pack_sent_fail);
		}

		if (strstr(buff, "received packets:") != NULL) {
			sscanf(buff, "received packets: %u", &pack_recv);
		}

		if (strstr(buff, "dropped packets:") != NULL) {
			sscanf(buff, "dropped packets: %u", &pack_drop);
		}
	}

	pclose(fp);
	add_counters_to_blob(bb, pack_sent, pack_sent_fail, pack_recv, pack_drop);
}

void get_current_status(const char *obj, const char *sec, struct blob_buf *bb)
{
	bool client_enable = false;

	if (!bb)
		return;

	if (!obj) {
		blobmsg_add_string(bb, "Status", "Error");
		return;
	}

	bool timemngr_enable = get_timemngr_enable();
	if (strcmp(obj, "server") == 0) {
		if (false == timemngr_enable) {
			blobmsg_add_string(bb, "Status", "Down");
			return;
		}

		if (false == get_timemngr_server_enable()) {
			blobmsg_add_string(bb, "Status", "Down");
			return;
		}

		if (false == ntpd_running()) {
			blobmsg_add_string(bb, "Status", "Error");
			return;
		}

		blobmsg_add_string(bb, "Status", "Up");
		return;
	} else if (strcmp(obj, "global") == 0) {
		if (false == timemngr_enable) {
			blobmsg_add_string(bb, "Status", "Disabled");
			return;
		}

		// Check if any of the client instance is enabled
		if (true == get_timemngr_client_enable(NULL))
			client_enable = true;
	} else if (strcmp(obj, "client") == 0) {
		if (sec == NULL || strlen(sec) == 0) {
			blobmsg_add_string(bb, "Status", "Error");
			return;
		}

		if (false == timemngr_enable) {
			blobmsg_add_string(bb, "Status", "Disabled");
			return;
		}

		// Check the client instance is enabled
		if (true == get_timemngr_client_enable(sec))
			client_enable = true;
	} else {
		blobmsg_add_string(bb, "Status", "Error");
		return;
	}

	if (true == client_enable) {
		if (false == ntpd_running()) {
			blobmsg_add_string(bb, "Status", "Error");
			return;
		}

		if (false == ntpd_leap_alarm()) {
			blobmsg_add_string(bb, "Status", "Synchronized");
			return;
		}

		blobmsg_add_string(bb, "Status", "Unsynchronized");
		return;
	}

	blobmsg_add_string(bb, "Status", "Disabled");
	return;
}

void get_current_stats(const char *obj, const char *sec, struct blob_buf *bb)
{
	if (!bb)
		return;

	if (!obj) {
		PRINT_ERR("Object name is NULL");
		add_counters_to_blob(bb, 0, 0, 0, 0);
		return;
	}

	if (false == get_timemngr_enable()) {
		PRINT_DEBUG("timemngr is globally disabled");
		add_counters_to_blob(bb, 0, 0, 0, 0);
		return;
	}

	if (false == ntpd_running()) {
		PRINT_DEBUG("NTPD is not running");
		add_counters_to_blob(bb, 0, 0, 0, 0);
		return;
	}

	if (strcmp(obj, "server") == 0) {
		if (false == get_timemngr_server_enable()) {
			PRINT_DEBUG("timemngr server instance is disabled");
			add_counters_to_blob(bb, 0, 0, 0, 0);
			return;
		}
	} else if (strcmp(obj, "client") == 0) {
		if (sec == NULL || strlen(sec) == 0) {
			PRINT_DEBUG("Specific client name is given");
			add_counters_to_blob(bb, 0, 0, 0, 0);
			return;
		}

		// Check the client instance is enabled
		if (false == get_timemngr_client_enable(sec)) {
			PRINT_DEBUG("timemngr client instance is disabled");
			add_counters_to_blob(bb, 0, 0, 0, 0);
			return;
		}
	} else {
		PRINT_ERR("Invalid object name: %s", obj);
		add_counters_to_blob(bb, 0, 0, 0, 0);
		return;
	}

	update_ntpd_counters(bb);
	return;
}

void print_log(int severity, const char *format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	vsyslog(severity, format, arglist);
	va_end(arglist);
}

void timemngr_schedule_sync(struct uloop_timeout *timeout __attribute__((unused)))
{
	if (false == ntpd_leap_alarm()) {
		uint32_t id;
		int rc = -1;
		struct blob_buf b;
		void *a;

		memset(&b, 0, sizeof(struct blob_buf));
		blob_buf_init(&b, 0);

		a = blobmsg_open_array(&b, "env");
		blobmsg_add_string(&b, NULL, "ACTION=stratum");
		blobmsg_close_array(&b, a);

		if (!ubus_lookup_id(g_ubus_ctx, "hotplug.ntp", &id))
			rc = ubus_invoke(g_ubus_ctx, id, "call", b.head, NULL, NULL, 5000);

		if (rc != 0) {
			blob_buf_free(&b);
			goto end;
		}

		blob_buf_free(&b);
		start_sync_timer(660);
		return;
	}

end:
	start_sync_timer(10);
}

void start_sync_timer(int seconds)
{
	uloop_timeout_set(&timemngr_sync_timer, seconds * 1000);
	PRINT_DEBUG("Sync timer called %ld", (long int)uloop_timeout_remaining64(&timemngr_sync_timer));
}
